from random import randint
from typing import Iterable
from itertools import groupby

def rand_n_digit_int(num_digits: int, base: int = 2):
    return randint(base**(num_digits-1), base**num_digits-1)

def all_equal(seq: Iterable):
    g = groupby(seq)
    return next(g, True) and not next(g, False)

def dot_to_and(s: str, bold: bool = False) -> str:
    if bold:
        t = s.split('.')
        return '\\textbf{%s}&\\textbf{%s}'%(t[0],t[1]) if '.' in s else '\\textbf{%s}&'%s
    else:
        return s.replace('.','&') if '.' in s else s+'&'


from random import randrange
from typing import Dict
from collections import Counter
from .small_primes import small_primes

try:
    from sympy.ntheory import isprime
except (ImportError, ModuleNotFoundError):
    from .primes import isprime

try:
    from gmpy2 import gcd, powmod
except ImportError:
    from math import gcd
    powmod = pow

def pollard_brent(n: int) -> int:
    """
    Given a positive integer n, returns a nontrivial factor of n if n is
    composite; if n is prime, returns n. Uses Pollard's rho algorithm with
    Brent's cycle finding method:
    https://comeoncodeon.wordpress.com/2010/09/18/pollard-rho-brent-integer-factorization/.

    Parameters
    ----------
    n: positive integer.

    Returns
    -------
        a nontrivial factor of n if n is composite; n otherwise.
    """
    if not n&1: # if n%2 == 0
        return 2
    elif not n%3: # if n%3 == 0
        return 3
    
    y, c, m = randrange(1, n), randrange(1, n), randrange(1, n)
    g, r, q = 1, 1, 1
    while g == 1:
        x = y
        for i in range(r):
            y = (powmod(y, 2, n) + c) % n
        k = 0
        while k < r and g == 1:
            ys = y
            for i in range(min(m, r-k)):
                y = (powmod(y, 2, n) + c) % n
                q = q*abs(x-y) % n
            g = gcd(q, n)
            k += m
        r *= 2
    if g == n:
        while True:
            ys = (powmod(ys, 2, n) + c) % n
            g = gcd(abs(x-ys), n)
            if g > 1:
                break
    return g

def large_factors(n: int) -> Dict[int, int]:
    """
    Given a positive integer n, returns a dict containing the prime factors of n
    as keys and their respective multiplicities as values.

    Parameters
    ----------
    n: positive integer.

    Returns
    -------
        a dict containing the prime factors of n as keys and their respective
        multiplicities as values.
    """
    factors = Counter()
    while n > 1:
        if isprime(n):
            factors[n] += 1
            break
        factor = pollard_brent(n)
        factors += large_factors(factor)
        n //= factor
    return factors

def factorint(n: int) -> Dict[int, int]:
    """
    Given a positive integer n, returns a dict containing the prime factors of n
    as keys and their respective multiplicities as values.

    Parameters
    ----------
    n: positive integer.

    Returns
    -------
        a dict containing the prime factors of n as keys and their respective
        multiplicities as values.
    """
    factors = Counter()
    for p in small_primes:
        while not n%p: # while n%p == 0
            factors[p] += 1
            n //= p
    if n == 1:
        return factors
    return factors + large_factors(n)


from random import randrange
from .small_primes import small_primes, max_prime

try:
    from gmpy2 import powmod
except ImportError:
    powmod = pow

def isprime(n: int, k: int = 3) -> bool:
    """
    Given an integer n and a positive integer k, determines if n is likely to be
    prime using the Miller-Rabin primality test. The probability of a false
    positive is 1/4ᵏ.

    Parameters
    ----------
    n: integer.
    k: positive integer, optional (default is 3)
        The number of rounds of testing to perform.
    
    Returns
    -------
        True if n is likely to be prime; False otherwise.
    """
    if n < 2:
        return False
    elif n < 4:
        return n > 1 # 2 and 3 are trivial primes
    elif not n&1: # if n%2 == 0
        return False
    elif n < max_prime:
        return n in small_primes
    
    d, r = n-1, 0 # n = 2ʳ·d + 1
    while not d&1: # while d%2 == 0
        d >>= 1
        r += 1
    
    for _ in range(k):
        x = powmod(randrange(2, n-1), d, n)
        if x == 1 or x == n-1:
            continue
        for _ in range(r-1):
            x = powmod(x, 2, n)
            if x == 1:
                return False
            elif x == n-1:
                break
        else:
            return False
    return True


from .factor import factorint

def totient(n: int) -> int:
    """
    Given a positive integer n, returns φ(n), which is the number of positive
    integers ≤ n that are coprime to n. Uses the product formula for φ: letting
        n = p₁ᵉ¹p₂ᵉ²···pᵣᵉʳ
    be the unique prime factorisation of n, where r is a nonnegative integer,
    p₁ < p₂ < ··· < pᵣ are distinct primes, and each eᵢ is a positive integer,
        φ(n) = n·(1 - 1/p₁)·(1 - 1/p₂)···(1 - 1/pᵣ).

    Parameters
    ----------
    n: positive integer.

    Returns
    -------
        φ(n).
    """
    for p in factorint(n).keys():
        n -= n//p
    return n